{
 Designer: Craig Ward, 100554.2072@compuserve.com
 Date:     20/2/96
 Version:  1.0


 Hacked by:Simon Armstrong, 100717.3531@compuserve.com
 Date:     24/2/96
 Version:  2.0

           * Derived from TCustomPanel instead of TPanel.
           * Remove csSetCaption from ControlStyle.
           * Do not assume contained controls are TSpeedButtons.
           * Do not assume fixed size for contained controls.
           * Allow either vertical or horizontal orientation when floating.  Double-
             click to toggle between the two.  Publish orientation property.
           * Changed a few names to distinguish from property set methods.


 Function: Floating Tool-Bar.

 Notes:    This component is a descendent of the TCustomPanel component. This descendent
           differs from its parent in that it has a handler attached to the OnMouseDown
           event, which initiates the dragging of the control.

           A single click will un-dock a docked toolbar.

           There are several custom methods that are initiated in response to that
           handler:

            [1] LayoutPanel - this checks to see if the toolbar has gone over the form
                boundaries, in which case it sets the alignment property
            [2] LayoutPanelButtons - this redraws the panel's buttons which is necessary
                if the panel is moving from horizontal layout to vertical.  Note that
                the code takes into account spaces between the controls, which is
                necessary if we are to keep the original relative "positions".

 Problems: [1] it won't wrap buttons to more than one line (yet!)
           [2] it does not draw border icons on the panel if it is in a floating state (ie:
               the panel will appear as a panel, and not as a mini child-window). If anyone knows
               how to do this, then please let me know, but at present I can't see how this
               will work unless I find a way to cheat (ie: make the panel look like a child-window).

 Credits:  Thanks to Mike Orriss (100570.121@compuserve.com) for pointing out the
           undocumented SC_DRAGMOVE parameter. That single undocumented parameter saved a huge
           amount of time and effort.

********************************************************************************}
unit cwTool;

interface

uses
  Buttons, Classes, Controls, ExtCtrls, Forms, Graphics, Messages, WinTypes;

type
  TOrientation = (orVertical, orHorizontal);

  TFloatingToolBar = class(TCustomPanel)
  private
    { Private declarations }
    FDocked: boolean;
    FOrientation: TOrientation;
    FPreviousOrientation: TOrientation;
    FCentred: boolean;
    FControls: TList;            { a copy of Controls[] sorted in to X,Y order }
    function  CalculateBorder: integer;
    procedure CopyAndSortControls;
    function  GetAlign: TAlign;
    function  GetBevelInner: TPanelBevel;
    function  GetBevelOuter: TPanelBevel;
    function  GetBevelWidth: TBevelWidth;
    function  GetBorderStyle: TBorderStyle;
    function  GetBorderWidth: TBorderWidth;
    procedure LayoutPanel;
    procedure LayoutPanelButtons;
    procedure LayoutPanelWhileHidden;
    procedure MouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure SetAlign(const a: TAlign);
    procedure SetBevelInner(const b: TPanelBevel);
    procedure SetBevelOuter(const b: TPanelBevel);
    procedure SetBevelWidth(const w: TBevelWidth);
    procedure SetBorderStyle(const s: TBorderStyle);
    procedure SetBorderWidth(const w: TBorderWidth);
    procedure SetCentre(const c: boolean);
    procedure SetOrientation(const o: TOrientation);
  protected
    { Protected declarations }
    procedure AlignControls(AControl: TControl; var sRect: TRect); override;
    procedure Loaded; override;
  public
    { Public declarations }
    constructor Create(aOwner: TComponent); override;
  published
    { Published declarations }
    property Align: TAlign read GetAlign write SetAlign;
    property BevelInner: TPanelBevel read GetBevelInner write SetBevelInner;
    property BevelOuter: TPanelBevel read GetBevelOuter write SetBevelOuter;
    property BevelWidth: TBevelWidth read GetBevelWidth write SetBevelWidth;
    property BorderStyle: TBorderStyle read GetBorderStyle write SetBorderStyle;
    property BorderWidth: TBorderWidth read GetBorderWidth write SetBorderWidth;
    property Centred: boolean read FCentred write SetCentre default true;
    property Color;
    property Ctl3D;
    property Enabled;
    property Font;
    property Orientation: TOrientation read FOrientation write SetOrientation default orHorizontal;
    property Locked;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property ShowHint;
    property Visible;
  end;

procedure Register;

implementation

uses
   WinProcs;

{register}
procedure Register;
begin
   RegisterComponents('Extra', [TFloatingToolBar]);
end;

{on create}
constructor TFloatingToolBar.Create(aOwner: TComponent);
begin
   inherited create(aOwner);
   ControlStyle := ControlStyle + [csAcceptsControls, csOpaque] - [csSetCaption];
   FOrientation := orHorizontal;
   FPreviousOrientation := FOrientation;
   OnMouseDown := MouseDown;
   FControls := TList.Create;
   FCentred := true;
end;

{on mouse down}
procedure TFloatingToolBar.MouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
const
   SC_DRAGMOVE : Longint = $F012;
var
   p: TPoint;
begin
   if button = mbLeft then begin
      if ssDouble in Shift then begin  { do it here (instead of DblClick) so can use mouse position }
         if not FDocked then begin
            Visible := false;          { prevent flickering while sort out new shape and position }
            try
               if FOrientation = orVertical then
                  SetOrientation(orHorizontal)
               else
                  SetOrientation(orVertical);

               { centre on mouse position }
               Left := Left + x - Width div 2;
               Top := Top + y - Height div 2;

               { but don't dock unless user tells us to }
               if top < 0 then
                  top := 0
               else if top + height > parent.clientHeight then
                  top := parent.clientHeight - height;    { what if not placed on a form ?}
               if left < 0 then
                  left := 0
               else if (width + left) > parent.clientWidth then
                  left := parent.clientWidth - width;
            finally
               Visible := true;
            end;
         end;
      end else begin
         {let the API perform its tricks}
         ReleaseCapture;
         if FDocked then
            LayoutPanel;                           { undock it & minimise size }
         SendMessage(Handle, WM_SYSCOMMAND, SC_DRAGMOVE, 0);
         LayoutPanelWhileHidden;
      end;
   end;
end;

{***custom routines*************************************************************}

procedure TFloatingToolBar.AlignControls(AControl: TControl; var sRect: TRect);
begin
   inherited AlignControls(AControl, sRect);
   if FControls.Count <> ControlCount then
      CopyAndSortControls;
end;

function TFloatingToolBar.CalculateBorder: integer;
begin
   result := BorderWidth;
   if BorderStyle = bsSingle then
      inc(result);
   if BevelInner <> bvNone then
      inc(result,BevelWidth);
   if BevelOuter <> bvNone then
      inc(result,BevelWidth);
end;

procedure TFloatingToolBar.CopyAndSortControls;
var
   i: integer;
   c1,c2: TControl;
   f: boolean;
begin
   FControls.Clear;
   { copy controls into a local list }
   for i := 0 to ControlCount - 1 do
      FControls.Add(Controls[i]);

   { sort the local list into increasing X,Y order }
   repeat
      f := false;
      for i := 0 to FControls.Count - 2 do begin
         c1 := FControls[i];
         c2 := FControls[i+1];
         if ((FOrientation = orHorizontal) and
             (c1.Left > c2.Left) or
            ((c1.Left = c2.Left) and
             (c1.Top  > c2.Top))) or
            ((FOrientation = orVertical) and
             (c1.Top  > c2.Top) or
            ((c1.Top  = c2.Top) and
             (c1.Left > c2.Left))) then begin
            f := true;
            FControls.Exchange(i,i+1);
         end;
      end;
   until not f;
end;

function TFloatingToolBar.GetAlign: TAlign;
begin
   result := inherited align;
end;

function TFloatingToolBar.GetBevelInner: TPanelBevel;
begin
   result := inherited BevelInner;
end;

function TFloatingToolBar.GetBevelOuter: TPanelBevel;
begin
   result := inherited BevelOuter;
end;

function TFloatingToolBar.GetBevelWidth: TBevelWidth;
begin
   result := inherited BevelWidth;
end;

function TFloatingToolBar.GetBorderStyle: TBorderStyle;
begin
   result := inherited BorderStyle;
end;

function TFloatingToolBar.GetBorderWidth: TBorderWidth;
begin
   result := inherited BorderWidth;
end;

{set the panel's position}
procedure TFloatingToolBar.LayoutPanel;
begin
   {if the panel is over the form's boundaries, then set alignment}
   if top < 0 then begin
      align := alTop;
      FDocked := true;
      SetOrientation(orHorizontal);
   end else if top + height > parent.clientHeight then begin
      align := alBottom;
      FDocked := true;
      SetOrientation(orHorizontal);
   end else if left < 0 then begin
      align := alLeft;
      FDocked := true;
      SetOrientation(orVertical);
   end else if width + left > parent.clientWidth then begin
      align := alRight;
      FDocked := true;
      SetOrientation(orVertical);
   end else begin                   { undocked, but leave orientation as it is }
      align := alNone;
      FDocked := false;
      LayoutPanelButtons;
   end;
end;

procedure TFloatingToolBar.LayoutPanelWhileHidden;
var
   v: boolean;
begin { prevent any change of form scroll-bars being shown momentarily during re-size }
   v := visible;
   Visible := false;
   try
      if csDesigning in ComponentState then
         LayoutPanelButtons          { don't want auto docking while designing }
      else
         LayoutPanel;
   finally
      Visible := v;
   end;
end;

{set the panel's buttons - this call is necessary if moving from horizontal to
 a vertical Orientation}
procedure TFloatingToolBar.LayoutPanelButtons;

   function CalculateSeparation(const index: integer): integer;
   var
      c1,c2: TControl;
   begin
      if index < (FControls.Count - 1) then begin
         c1 := fcontrols[index];
         c2 := fcontrols[index+1];
         if FPreviousOrientation = orVertical then
            result := c2.top - c1.top - c1.height
         else
            result := c2.left - c1.left - c1.width;
      end else
         result := 0;
   end;

   var
   i, y, x, b, d: integer;
   Max: integer;
   c1,c2: TControl;
begin
   if FControls.Count = 0 then
      exit;

   case FOrientation of
      orHorizontal:
         begin
            max := 0;
            b := CalculateBorder;
            x := b - 1;
            for i := 0 to FControls.Count - 1 do begin
               d := CalculateSeparation(i);                 { before moving it!}
               c1 := fcontrols[i];
               c1.left := x;
               c1.top := b - 1;
               x := x + c1.width + d;
               if c1.height > max then
                  max := c1.height;
            end;
            width := x + b;
            height := 2 * b + max;
            if FCentred then
               for i := 0 to FControls.Count - 1 do begin
                  c1 := FControls[i];
                  c1.Top := (Height - c1.Height) div 2;
               end;
         end;

      orVertical:
         begin
            max := 0;
            b := CalculateBorder;
            y := b - 1;
            for i := 0 to FControls.Count - 1 do begin
               d := CalculateSeparation(i);                 { before moving it!}
               c1 := fcontrols[i];
               c1.left := b - 1;
               c1.top := y;
               y := y + c1.Height + d;
               if c1.width > max then
                  max := c1.width;
            end;
            width := 2 * b + max;
            height := y + b;
            if FCentred then
               for i := 0 to FControls.Count - 1 do begin
                  c1 := FControls[i];
                  c1.left := (Width - c1.width) div 2;
               end;
         end;
   end;
   FPreviousOrientation := FOrientation;
end;

procedure TFloatingToolBar.Loaded;
begin
   inherited loaded;
   CopyAndSortControls;
   FPreviousOrientation := FOrientation;
   LayoutPanel;
end;

procedure TFloatingToolBar.SetAlign(const a: TAlign);
begin
   if inherited align <> a then begin
      if (a = alTop) or (a = alBottom) then
         SetOrientation(orHorizontal)
      else if (a = alLeft) or (a = alRight) then
         SetOrientation(orVertical);
      inherited align := a;
   end;
end;

procedure TFloatingToolBar.SetBevelInner(const b: TPanelBevel);
begin
   if inherited BevelInner <> b then begin
      inherited BevelInner := b;
      LayoutPanelWhileHidden;
   end;
end;

procedure TFloatingToolBar.SetBevelOuter(const b: TPanelBevel);
begin
   if inherited BevelOuter <> b then begin
      inherited BevelOuter := b;
      LayoutPanelWhileHidden;
   end;
end;

procedure TFloatingToolBar.SetBevelWidth(const w: TBevelWidth);
begin
   if inherited BevelWidth <> w then begin
      inherited BevelWidth := w;
      LayoutPanelWhileHidden;
   end;
end;

procedure TFloatingToolBar.SetBorderStyle(const s: TBorderStyle);
begin
   if inherited BorderStyle <> s then begin
      inherited BorderStyle := s;
      LayoutPanelWhileHidden;
   end;
end;

procedure TFloatingToolBar.SetBorderWidth(const w: TBorderWidth);
begin
   if inherited BorderWidth <> w then begin
      inherited BorderWidth := w;
      LayoutPanelWhileHidden;
   end;
end;

procedure TFloatingToolBar.SetCentre(const c: boolean);
begin
   if FCentred <> c then begin
      FCentred := c;
      LayoutPanelButtons;
   end;
end;

procedure TFloatingToolBar.SetOrientation(const o: TOrientation);
begin
   if FOrientation <> o then begin
      FPreviousOrientation := FOrientation;
      FOrientation := o;
      inherited Align := alNone;
      LayoutPanelButtons;
   end;
end;

end.
